from typing import List, Optional, Sequence, Callable, overload
from fastapi.params import Depends
from fastapi.types import DecoratedCallable

import functools
import functions_framework
import asyncio
import flask

from ._proxy import mount_entry_point, proxy


@overload
def handler(__fn: DecoratedCallable) -> DecoratedCallable:
    ...


@overload
def handler(
    *,
    dependencies: Optional[Sequence[Depends]] = None,
    methods: Optional[List[str]] = None,
) -> Callable[[DecoratedCallable], DecoratedCallable]:
    ...


def handler(
    __fn: Optional[DecoratedCallable] = None,
    *,
    dependencies: Optional[Sequence[Depends]] = None,
    methods: Optional[List[str]] = None,
):
    """
    Decorator for creating HTTP handlers that can be used as entry points.

    This decorator can be used to create HTTP handlers that will be treated as entry points
    in the Google Cloud Function acting like FastAPI handler.

    Args:
        dependencies (Optional[Sequence[Depends]], optional): A list of FastAPI dependencies.
            Defaults to None.
        methods (Optional[List[str]], optional): A list of HTTP methods that this handler should respond to.
            Defaults to None.

    Returns:
        Callable[[DecoratedCallable], DecoratedCallable]] | DecoratedCallable: A decorator function that can be applied to a handler function.
    """

    def decorator(func: DecoratedCallable) -> DecoratedCallable:
        mount_entry_point(func, dependencies=dependencies, methods=methods)
        func = functions_framework.http(func)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return asyncio.run(proxy(flask.request))

        return wrapper

    return decorator(__fn) if __fn else decorator



import asyncio
from typing import Generator, AsyncGenerator, Iterator, AsyncIterator, Any, TypeVar
from httpx._types import AsyncByteStream, SyncByteStream


T = TypeVar("T")


def sync_iterator_to_sync_generator(sync_iterator: Iterator[T]) -> Generator[T, None, None]:
    for i in sync_iterator:
        yield i


async def async_iterator_to_async_generator(async_iterator: AsyncIterator[T]) -> AsyncGenerator[T, None]:
    async for i in async_iterator:
        yield i


async def sync_iterator_to_async_generator(sync_iterator: Iterator[T]) -> AsyncGenerator[T, None]:
    for i in sync_iterator:
        yield i


def async_iterator_to_sync_generator(async_iterator: AsyncIterator[T]) -> Generator[T, None, None]:
    while True:
        try:
            yield asyncio.run(async_iterator.__anext__())
        except StopAsyncIteration:
            break


class StreamExtensionsBase():
    @classmethod
    def from_sync_iterator(cls, iterator: Iterator[bytes]) -> Any:
        raise NotImplementedError()

    @classmethod
    def from_async_iterator(cls, iterator: AsyncIterator[bytes]) -> Any:
        raise NotImplementedError()


class ExtendedAsyncByteStream(AsyncByteStream, StreamExtensionsBase):
    def __init__(
        self, generator: AsyncGenerator[bytes, None]
    ) -> None:
        self._generator = generator

    def __aiter__(self) -> AsyncIterator[bytes]:
        return self._generator.__aiter__()
    
    async def aclose(self) -> None:
        await self._generator.aclose()
    
    @classmethod
    def from_async_iterator(cls, iterator: AsyncIterator[bytes]) -> "ExtendedAsyncByteStream":
        async_generator = async_iterator_to_async_generator(iterator)
        return cls(async_generator)
    
    @classmethod
    def from_sync_iterator(cls, iterator: Iterator[bytes]) -> "ExtendedAsyncByteStream":
        async_generator = sync_iterator_to_async_generator(iterator)
        return cls(async_generator)


class ExtendedSyncByteStream(SyncByteStream, StreamExtensionsBase):
    def __init__(
        self, generator: Generator[bytes, None, None]
    ) -> None:
        self._generator = generator

    def __iter__(self) -> Iterator[bytes]:
        return self._generator.__iter__()

    def close(self) -> None:
        self._generator.close()
    
    @classmethod
    def from_async_iterator(cls, iterator: AsyncIterator[bytes]) -> "ExtendedSyncByteStream":
        sync_generator = async_iterator_to_sync_generator(iterator)
        return cls(sync_generator)
    
    @classmethod
    def from_sync_iterator(cls, iterator: Iterator[bytes]) -> "ExtendedSyncByteStream":
        sync_generator = sync_iterator_to_sync_generator(iterator)
        return cls(sync_generator)



from typing import Callable, Optional, Sequence, List, overload
from fastapi.params import Depends
from fastapi.types import DecoratedCallable
from ._handler import handler


class router:
    """
    Unlock the full potential of Google Cloud Functions with FastAPI and async capabilities using the 'router' class!

    Are you looking to create a single powerful entry point for your Google Cloud Function?
    The 'router' class is designed for just that, providing seamless integration with FastAPI and async code.

    Key Highlights:
    - Easily register a single entry point for your Google Cloud Function.
    - Leverage FastAPI for robust request handling and request validation.
    - Harness the power of async code for responsive and scalable functions.
    - Define your function's input parameters with ease.

    Example Usage:

    ```python
    # Import necessary modules
    import asyncio
    from datetime import date
    from fastgcf import router

    # Simply use a decorator
    @router.get
    async def main(start_date: date, end_date: date):
        await asyncio.sleep(1)  # Simulate async processing
        return {"start_date": start_date, "end_date": end_date}

    # That's it! Your Google Cloud Function is ready to handle and validate async GET requests seamlessly.
    ```
    """


    @staticmethod
    @overload
    def get(__fn: DecoratedCallable) -> DecoratedCallable:
        ...

    @staticmethod
    @overload
    def get(
        *,
        dependencies: Optional[Sequence[Depends]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        ...

    @staticmethod
    def get(
        __fn: Optional[DecoratedCallable] = None,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
    ):
        return handler(__fn, dependencies=dependencies, methods=['GET'])


    @staticmethod
    @overload
    def put(__fn: DecoratedCallable) -> DecoratedCallable:
        ...

    @staticmethod
    @overload
    def put(
        *,
        dependencies: Optional[Sequence[Depends]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        ...

    @staticmethod
    def put(
        __fn: Optional[DecoratedCallable] = None,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
    ):
        return handler(__fn, dependencies=dependencies, methods=['PUT'])
    

    @staticmethod
    @overload
    def post(__fn: DecoratedCallable) -> DecoratedCallable:
        ...

    @staticmethod
    @overload
    def post(
        *,
        dependencies: Optional[Sequence[Depends]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        ...

    @staticmethod
    def post(
        __fn: Optional[DecoratedCallable] = None,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
    ):
        return handler(__fn, dependencies=dependencies, methods=['POST'])


    @staticmethod
    @overload
    def options(__fn: DecoratedCallable) -> DecoratedCallable:
        ...

    @staticmethod
    @overload
    def options(
        *,
        dependencies: Optional[Sequence[Depends]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        ...

    @staticmethod
    def options(
        __fn: Optional[DecoratedCallable] = None,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
    ):
        return handler(__fn, dependencies=dependencies, methods=['OPTIONS'])
    

    @staticmethod
    @overload
    def head(__fn: DecoratedCallable) -> DecoratedCallable:
        ...

    @staticmethod
    @overload
    def head(
        *,
        dependencies: Optional[Sequence[Depends]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        ...

    @staticmethod
    def head(
        __fn: Optional[DecoratedCallable] = None,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
    ):
        return handler(__fn, dependencies=dependencies, methods=['HEAD'])


    @staticmethod
    @overload
    def patch(__fn: DecoratedCallable) -> DecoratedCallable:
        ...

    @staticmethod
    @overload
    def patch(
        *,
        dependencies: Optional[Sequence[Depends]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        ...

    @staticmethod
    def patch(
        __fn: Optional[DecoratedCallable] = None,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
    ):
        return handler(__fn, dependencies=dependencies, methods=['PATCH'])


    @staticmethod
    @overload
    def route(__fn: DecoratedCallable) -> DecoratedCallable:
        ...

    @staticmethod
    @overload
    def route(
        *,
        dependencies: Optional[Sequence[Depends]] = None,
        methods: Optional[List[str]] = None,
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        ...

    @staticmethod
    def route(
        __fn: Optional[DecoratedCallable] = None,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
        methods: Optional[List[str]] = None,
    ):
        return handler(__fn, dependencies=dependencies, methods=methods)



import flask
import httpx
from ._streams import ExtendedSyncByteStream


def convert_flask_request_to_httpx_request(request: flask.Request) -> httpx.Request:
    """
    Convert a Flask HTTP request to an HTTPX request.

    Args:
        request (flask.Request): The incoming Flask HTTP request.

    Returns:
        httpx.Request: An HTTPX request.
    """

    method = request.method
    url = request.url
    form_data = request.form.to_dict(flat=False) if len(request.form) > 0 else None
    headers = {key.lower(): value for key, value in request.headers}
    data = request.data if form_data is None else form_data

    files = []
    for file_name, file_stream in request.files.items(multi=True):
        file_headers = {key: value for key, value in file_stream.headers.items()}
        files.append((file_name, (file_stream.filename, file_stream.stream, file_stream.mimetype, file_headers)))

    has_files = len(files) > 0

    if len(data) == 0:
        data = None

    if not has_files:
        files = None

    headers.pop('content-length', None)

    httpx_request = httpx.Request(
        method=method,
        url=url,
        headers=headers,
        data=data,
        files=files,
    )

    # Ensure that the request is encoded as 'application/x-www-form-urlencoded'
    # when form data is present and the 'files' object is falsy,
    # because HTTPX doesn't treat such requests as 'multipart/form-data'
    # even if the content type is set beforehand.
    if not has_files and form_data is not None:
        httpx_request.headers['content-type'] = 'application/x-www-form-urlencoded'

    return httpx_request


def convert_httpx_response_to_flask_response(response: httpx.Response) -> flask.Response:
    """
    Convert an HTTPX response to a Flask response.

    Args:
        response (httpx.Response): The HTTPX response.

    Returns:
        flask.Response: A Flask response.
    """

    response_stream = ExtendedSyncByteStream.from_async_iterator(response.aiter_bytes())
    flask_response = flask.Response(response_stream, status=response.status_code)

    for header, value in response.headers.items():
        flask_response.headers[header] = value

    return flask_response



from setuptools import setup, find_packages
from pathlib import Path
import re


NAME = 'fastgcf'
DESCRIPTION = 'FastAPI in Google Cloud Functions'
AUTHOR = 'Tigran Zalian'
AUTHOR_EMAIL = 'tigraanzalian@gmail.com'
URL = 'https://github.com/TigranZalian/fastgcf'

package_dir = Path(__file__).parent / NAME
package_init_file = package_dir / '__init__.py'

VERSION = re.search(r"__version__ = '(.*)'", package_init_file.read_text()).group(1)

with open('./requirements.txt', 'r') as f:
    INSTALL_REQUIRES = list(filter(lambda l: bool(l.strip()), f.readlines()))

PACKAGES = find_packages()

CLASSIFIERS = [
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
]

with open('README.md', 'r', encoding='utf-8') as readme_file:
    LONG_DESCRIPTION = readme_file.read()

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    packages=PACKAGES,
    install_requires=INSTALL_REQUIRES,
    classifiers=CLASSIFIERS,
)



import fastapi
import flask
import httpx
import nest_asyncio

from fastapi.params import Depends
from typing import Any, Callable, List, Optional, Sequence

from ._transport import ASGITransport
from ._converters import convert_flask_request_to_httpx_request, convert_httpx_response_to_flask_response


app = fastapi.FastAPI(docs_url=None, redoc_url=None, openapi_url=None)
is_entry_point_mounted = False
asgi_transport = ASGITransport(app)
async_client = httpx.AsyncClient(transport=asgi_transport, base_url='http://server')


async def proxy(flask_request: flask.Request) -> flask.Response:
    """
    Proxy an incoming Flask request to the current FastAPI application.

    Args:
        flask_request (flask.Request): The incoming Flask request.

    Returns:
        flask.Response: The Flask response from the proxy operation.
    """

    httpx_request = convert_flask_request_to_httpx_request(flask_request)
    httpx_response = await async_client.send(httpx_request, stream=True)
    flask_response = convert_httpx_response_to_flask_response(httpx_response)
    return flask_response


def mount_entry_point(
    endpoint: Callable[..., Any],
    *,
    dependencies: Optional[Sequence[Depends]] = None,
    methods: Optional[List[str]] = None,
):
    """
    Mounts the entry point for the FastAPI application.

    This function registers the provided `endpoint` as the entry point for the FastAPI application.

    Args:
        endpoint (Callable[..., Any]): The function to be used as the entry point.
        dependencies (Optional[Sequence[Depends]], optional): A list of FastAPI dependencies.
            Defaults to None.
        methods (Optional[List[str]], optional): A list of HTTP methods that this entry point should respond to.
            Defaults to None.

    Raises:
        RuntimeError: If the entry point is already mounted.

    Note:
        This function should typically be called only once to set up the main entry point for the application.

    Example:
        mount_entry_point(my_endpoint_function, dependencies=[get_token])
    """

    global is_entry_point_mounted

    if is_entry_point_mounted:
        raise RuntimeError('Entry point is already mounted')

    nest_asyncio.apply()

    app.add_api_route('/', endpoint, dependencies=dependencies, methods=methods)
    is_entry_point_mounted = True
